#include "../../UI/FileManager/resource.rc"
#include "../../UI/GUI/resource2.rc"

STRINGTABLE
BEGIN
  100 "7z zip rar 001 cab iso xz txz lzma tar cpio bz2 bzip2 tbz2 tbz gz gzip tgz tpz z taz lzh lha rpm deb arj vhd wim swm esd fat ntfs dmg hfs xar squashfs"
END
