#include "PasswordDialogRes.h"
#include "../../GuiCommon.rc"

#define xc 140
#define yc 72

IDD_PASSWORD  DIALOG  0, 0, xs, ys  MY_MODAL_DIALOG_STYLE  MY_FONT
CAPTION "Enter password"
BEGIN
  LTEXT    "&Enter password:", IDT_PASSWORD_ENTER, m, m, xc, 8
  EDITTEXT  IDE_PASSWORD_PASSWORD, m, 20, xc, 14, ES_PASSWORD | ES_AUTOHSCROLL
  CONTROL  "&Show password", IDX_PASSWORD_SHOW, MY_CHECKBOX, m, 42, xc, 10
  OK_CANCEL
END
