#include "PropertyNameRes.h"

LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US

STRINGTABLE
BEGIN
  IDS_PROP_PATH         "Path"
  IDS_PROP_NAME         "Name"
  IDS_PROP_EXTENSION    "Extension"
  IDS_PROP_IS_FOLDER    "Folder"
  IDS_PROP_SIZE         "Size"
  IDS_PROP_PACKED_SIZE  "Packed Size"
  IDS_PROP_ATTRIBUTES   "Attributes"
  IDS_PROP_CTIME        "Created"
  IDS_PROP_ATIME        "Accessed"
  IDS_PROP_MTIME        "Modified"
  IDS_PROP_SOLID        "Solid"
  IDS_PROP_C0MMENTED    "Commented"
  IDS_PROP_ENCRYPTED    "Encrypted"
  IDS_PROP_SPLIT_BEFORE "Split Before"
  IDS_PROP_SPLIT_AFTER  "Split After"
  IDS_PROP_DICTIONARY_SIZE "Dictionary"
  IDS_PROP_CRC          "CRC"
  IDS_PROP_FILE_TYPE    "Type"
  IDS_PROP_ANTI         "Anti"
  IDS_PROP_METHOD       "Method"
  IDS_PROP_HOST_OS      "Host OS"
  IDS_PROP_FILE_SYSTEM  "File System"
  IDS_PROP_USER         "User"
  IDS_PROP_GROUP        "Group"
  IDS_PROP_BLOCK        "Block"
  IDS_PROP_COMMENT      "Comment"
  IDS_PROP_POSITION     "Position"
  IDS_PROP_PREFIX       "Path Prefix"
  IDS_PROP_FOLDERS      "Folders"
  IDS_PROP_FILES        "Files"
  IDS_PROP_VERSION      "Version"
  IDS_PROP_VOLUME       "Volume"
  IDS_PROP_IS_VOLUME    "Multivolume"
  IDS_PROP_OFFSET       "Offset"
  IDS_PROP_LINKS        "Links"
  IDS_PROP_NUM_BLOCKS   "Blocks"
  IDS_PROP_NUM_VOLUMES  "Volumes"

  IDS_PROP_BIT64        "64-bit"
  IDS_PROP_BIG_ENDIAN   "Big-endian"
  IDS_PROP_CPU          "CPU"
  IDS_PROP_PHY_SIZE     "Physical Size"
  IDS_PROP_HEADERS_SIZE "Headers Size"
  IDS_PROP_CHECKSUM     "Checksum"
  IDS_PROP_CHARACTS     "Characteristics"
  IDS_PROP_VA           "Virtual Address"
  IDS_PROP_ID           "ID"
  IDS_PROP_SHORT_NAME   "Short Name"
  IDS_PROP_CREATOR_APP  "Creator Application"
  IDS_PROP_SECTOR_SIZE  "Sector Size"
  IDS_PROP_POSIX_ATTRIB "Mode"
  IDS_PROP_SYM_LINK     "Symbolic Link"
  IDS_PROP_ERROR        "Error"
  IDS_PROP_TOTAL_SIZE   "Total Size"
  IDS_PROP_FREE_SPACE   "Free Space"
  IDS_PROP_CLUSTER_SIZE "Cluster Size"
  IDS_PROP_VOLUME_NAME  "Label"
  IDS_PROP_LOCAL_NAME   "Local Name"
  IDS_PROP_PROVIDER     "Provider"
  IDS_PROP_NT_SECURITY  "NT Security"
  IDS_PROP_ALT_STREAM   "Alternate Stream"
  IDS_PROP_AUX          "Aux"
  IDS_PROP_DELETED      "Deleted"
  IDS_PROP_IS_TREE      "Is Tree"
  IDS_PROP_SHA1         "SHA-1"
  IDS_PROP_SHA256       "SHA-256"
  IDS_PROP_ERROR_TYPE   "Error Type"
  IDS_PROP_NUM_ERRORS   "Errors"
  IDS_PROP_ERROR_FLAGS  "Errors"
  IDS_PROP_WARNING_FLAGS "Warnings"
  IDS_PROP_WARNING      "Warning"
  IDS_PROP_NUM_STREAMS  "Streams"
  IDS_PROP_NUM_ALT_STREAMS "Alternate Streams"
  IDS_PROP_ALT_STREAMS_SIZE "Alternate Streams Size"
  IDS_PROP_VIRTUAL_SIZE "Virtual Size"
  IDS_PROP_UNPACK_SIZE  "Unpack Size"
  IDS_PROP_TOTAL_PHY_SIZE "Total Physical Size"
  IDS_PROP_VOLUME_INDEX "Volume Index"
  IDS_PROP_SUBTYPE      "SubType"
  IDS_PROP_SHORT_COMMENT "Short Comment"
  IDS_PROP_CODE_PAGE    "Code Page"
  IDS_PROP_IS_NOT_ARC_TYPE  "Is not archive type"
  IDS_PROP_PHY_SIZE_CANT_BE_DETECTED "Physical Size can't be detected"
  IDS_PROP_ZEROS_TAIL_IS_ALLOWED "Zeros Tail Is Allowed"
  IDS_PROP_TAIL_SIZE "Tail Size"
  IDS_PROP_EMB_STUB_SIZE "Embedded Stub Size"
  IDS_PROP_NT_REPARSE   "Link"
  IDS_PROP_HARD_LINK    "Hard Link"
  IDS_PROP_INODE        "iNode"
  IDS_PROP_STREAM_ID    "Stream ID"
  IDS_PROP_READ_ONLY    "Read-only"
  IDS_PROP_OUT_NAME     "Out Name"
  IDS_PROP_COPY_LINK    "Copy Link"
END
