

#define IDS_PROP_PATH             1003
#define IDS_PROP_NAME             1004
#define IDS_PROP_EXTENSION        1005
#define IDS_PROP_IS_FOLDER        1006
#define IDS_PROP_SIZE             1007
#define IDS_PROP_PACKED_SIZE      1008
#define IDS_PROP_ATTRIBUTES       1009
#define IDS_PROP_CTIME            1010
#define IDS_PROP_ATIME            1011
#define IDS_PROP_MTIME            1012
#define IDS_PROP_SOLID            1013
#define IDS_PROP_C0MMENTED        1014
#define IDS_PROP_ENCRYPTED        1015
#define IDS_PROP_SPLIT_BEFORE     1016
#define IDS_PROP_SPLIT_AFTER      1017
#define IDS_PROP_DICTIONARY_SIZE  1018
#define IDS_PROP_CRC              1019
#define IDS_PROP_FILE_TYPE        1020
#define IDS_PROP_ANTI             1021
#define IDS_PROP_METHOD           1022
#define IDS_PROP_HOST_OS          1023
#define IDS_PROP_FILE_SYSTEM      1024
#define IDS_PROP_USER             1025
#define IDS_PROP_GROUP            1026
#define IDS_PROP_BLOCK            1027
#define IDS_PROP_COMMENT          1028
#define IDS_PROP_POSITION         1029
#define IDS_PROP_PREFIX           1030
#define IDS_PROP_FOLDERS          1031
#define IDS_PROP_FILES            1032
#define IDS_PROP_VERSION          1033
#define IDS_PROP_VOLUME           1034
#define IDS_PROP_IS_VOLUME        1035
#define IDS_PROP_OFFSET           1036
#define IDS_PROP_LINKS            1037
#define IDS_PROP_NUM_BLOCKS       1038
#define IDS_PROP_NUM_VOLUMES      1039

#define IDS_PROP_BIT64            1041
#define IDS_PROP_BIG_ENDIAN       1042
#define IDS_PROP_CPU              1043
#define IDS_PROP_PHY_SIZE         1044
#define IDS_PROP_HEADERS_SIZE     1045
#define IDS_PROP_CHECKSUM         1046
#define IDS_PROP_CHARACTS         1047
#define IDS_PROP_VA               1048
#define IDS_PROP_ID               1049
#define IDS_PROP_SHORT_NAME       1050
#define IDS_PROP_CREATOR_APP      1051
#define IDS_PROP_SECTOR_SIZE      1052
#define IDS_PROP_POSIX_ATTRIB     1053
#define IDS_PROP_SYM_LINK         1054
#define IDS_PROP_ERROR            1055
#define IDS_PROP_TOTAL_SIZE       1056
#define IDS_PROP_FREE_SPACE       1057
#define IDS_PROP_CLUSTER_SIZE     1058
#define IDS_PROP_VOLUME_NAME      1059
#define IDS_PROP_LOCAL_NAME       1060
#define IDS_PROP_PROVIDER         1061
#define IDS_PROP_NT_SECURITY      1062
#define IDS_PROP_ALT_STREAM       1063
#define IDS_PROP_AUX              1064
#define IDS_PROP_DELETED          1065
#define IDS_PROP_IS_TREE          1066
#define IDS_PROP_SHA1             1067
#define IDS_PROP_SHA256           1068
#define IDS_PROP_ERROR_TYPE       1069
#define IDS_PROP_NUM_ERRORS       1070
#define IDS_PROP_ERROR_FLAGS      1071
#define IDS_PROP_WARNING_FLAGS    1072
#define IDS_PROP_WARNING          1073
#define IDS_PROP_NUM_STREAMS      1074
#define IDS_PROP_NUM_ALT_STREAMS  1075
#define IDS_PROP_ALT_STREAMS_SIZE 1076
#define IDS_PROP_VIRTUAL_SIZE     1077
#define IDS_PROP_UNPACK_SIZE      1078
#define IDS_PROP_TOTAL_PHY_SIZE   1079
#define IDS_PROP_VOLUME_INDEX     1080
#define IDS_PROP_SUBTYPE          1081
#define IDS_PROP_SHORT_COMMENT    1082
#define IDS_PROP_CODE_PAGE        1083
#define IDS_PROP_IS_NOT_ARC_TYPE  1084
#define IDS_PROP_PHY_SIZE_CANT_BE_DETECTED 1085
#define IDS_PROP_ZEROS_TAIL_IS_ALLOWED 1086
#define IDS_PROP_TAIL_SIZE        1087
#define IDS_PROP_EMB_STUB_SIZE    1088
#define IDS_PROP_NT_REPARSE       1089
#define IDS_PROP_HARD_LINK        1090
#define IDS_PROP_INODE            1091
#define IDS_PROP_STREAM_ID        1092
#define IDS_PROP_READ_ONLY        1093
#define IDS_PROP_OUT_NAME         1094
#define IDS_PROP_COPY_LINK        1095
