#include "LinkDialogRes.h"
#include "../../GuiCommon.rc"

#define xc 288
#define yc 214

#undef xRadioSize 
#define xRadioSize xc - m - 2

IDD_LINK  DIALOG  0, 0, xs, ys  MY_MODAL_RESIZE_DIALOG_STYLE  MY_FONT
CAPTION "Link"
BEGIN
  LTEXT       "Link from:", IDT_LINK_PATH_FROM, m, m, xc, 8
  COMBOBOX    IDC_LINK_PATH_FROM, m, 20, xc - bxsDots - m, 64, MY_COMBO_WITH_EDIT
  PUSHBUTTON  "...", IDB_LINK_PATH_FROM, xs - m - bxsDots, 18, bxsDots, bys, WS_GROUP

  LTEXT       "Link to:", IDT_LINK_PATH_TO, m, 48, xc, 8
  COMBOBOX    IDC_LINK_PATH_TO, m, 60, xc - bxsDots - m, 64, MY_COMBO_WITH_EDIT
  PUSHBUTTON  "...", IDB_LINK_PATH_TO, xs - m - bxsDots, 58, bxsDots, bys, WS_GROUP

  LTEXT       "", IDT_LINK_PATH_TO_CUR, m, 78, xc, 8

  GROUPBOX    "Link Type", IDG_LINK_TYPE, m, 104, xc, 90

  CONTROL    "Hard Link", IDR_LINK_TYPE_HARD, "Button", BS_AUTORADIOBUTTON | WS_GROUP,
             m + m, 120, xRadioSize, 10
  CONTROL    "File Symbolic Link", IDR_LINK_TYPE_SYM_FILE, "Button", BS_AUTORADIOBUTTON,
             m + m, 134, xRadioSize, 10
  CONTROL    "Directory Symbolic Link", IDR_LINK_TYPE_SYM_DIR, "Button", BS_AUTORADIOBUTTON,
             m + m, 148, xRadioSize, 10
  CONTROL    "Directory Junction", IDR_LINK_TYPE_JUNCTION, "Button", BS_AUTORADIOBUTTON,
             m + m, 162, xRadioSize, 10
  CONTROL    "WSL", IDR_LINK_TYPE_WSL, "Button", BS_AUTORADIOBUTTON,
             m + m, 176, xRadioSize, 10

  DEFPUSHBUTTON  "Link",   IDB_LINK_LINK, bx2, by, bxs, bys
  PUSHBUTTON     "Cancel", IDCANCEL,      bx1, by, bxs, bys
END
