#include "MessagesDialogRes.h"
#include "../../GuiCommon.rc"

#define xc 440
#define yc 160

IDD_MESSAGES  DIALOG  0, 0, xs, ys  MY_MODAL_RESIZE_DIALOG_STYLE  MY_FONT
CAPTION "7-Zip: Diagnostic messages"
{
  DEFPUSHBUTTON "&Close", IDOK, bx, by, bxs, bys
  CONTROL "List1", IDL_MESSAGE, "SysListView32",
          LVS_REPORT | LVS_SHOWSELALWAYS | LVS_NOSORTHEADER | WS_BORDER | WS_TABSTOP,
          m, m, xc, yc - bys - m
}
