#define IDD_SETTINGS     2500
#define IDD_SETTINGS_2  12500

#define IDX_SETTINGS_SHOW_DOTS              2501
#define IDX_SETTINGS_SHOW_REAL_FILE_ICONS   2502
#define IDX_SETTINGS_SHOW_SYSTEM_MENU       2503
#define IDX_SETTINGS_FULL_ROW               2504
#define IDX_SETTINGS_SHOW_GRID              2505
#define IDX_SETTINGS_SINGLE_CLICK           2506
#define IDX_SETTINGS_ALTERNATIVE_SELECTION  2507
#define IDX_SETTINGS_LARGE_PAGES            2508
