#define IDD_LINK  7700

#define IDB_LINK_LINK           7701

#define IDT_LINK_PATH_FROM      7702
#define IDT_LINK_PATH_TO        7703

#define IDG_LINK_TYPE           7710
#define IDR_LINK_TYPE_HARD      7711
#define IDR_LINK_TYPE_SYM_FILE  7712
#define IDR_LINK_TYPE_SYM_DIR   7713
#define IDR_LINK_TYPE_JUNCTION  7714
#define IDR_LINK_TYPE_WSL       7715


#define IDC_LINK_PATH_FROM    100
#define IDC_LINK_PATH_TO      101

#define IDT_LINK_PATH_TO_CUR  102

#define IDB_LINK_PATH_FROM    103
#define IDB_LINK_PATH_TO      104
