#define IDS_PROGRESS_REMOVE     3305

#define IDS_PROGRESS_ADD        3320
#define IDS_PROGRESS_UPDATE     3321
#define IDS_PROGRESS_ANALYZE    3322
#define IDS_PROGRESS_REPLICATE  3323
#define IDS_PROGRESS_REPACK     3324

#define IDS_PROGRESS_DELETE     3326
#define IDS_PROGRESS_HEADER     3327
