#define IDD_COMPRESS     4000
#define IDD_COMPRESS_2  14000

#define IDC_COMPRESS_ARCHIVE             100
#define IDB_COMPRESS_SET_ARCHIVE         101
#define IDC_COMPRESS_LEVEL               102
#define IDC_COMPRESS_UPDATE_MODE         103
#define IDC_COMPRESS_FORMAT              104
#define IDC_COMPRESS_VOLUME              105
#define IDC_COMPRESS_METHOD              106
#define IDC_COMPRESS_DICTIONARY          107
#define IDC_COMPRESS_ORDER               108
#define IDC_COMPRESS_SOLID               109
#define IDC_COMPRESS_THREADS             110
#define IDE_COMPRESS_PARAMETERS          111

#define IDT_COMPRESS_HARDWARE_THREADS    112
#define IDT_COMPRESS_MEMORY_VALUE        113
#define IDT_COMPRESS_MEMORY_DE_VALUE     114

#define IDG_COMPRESS_NTFS                115
#define IDC_COMPRESS_PATH_MODE           116

#define IDE_COMPRESS_PASSWORD1           120
#define IDE_COMPRESS_PASSWORD2           121
#define IDC_COMPRESS_ENCRYPTION_METHOD   122

#define IDT_COMPRESS_ARCHIVE_FOLDER      130

#define IDT_COMPRESS_PATH_MODE          3410

#define IDT_PASSWORD_ENTER              3801
#define IDT_PASSWORD_REENTER            3802
#define IDX_PASSWORD_SHOW               3803
#define IDS_PASSWORD_NOT_MATCH          3804
#define IDS_PASSWORD_USE_ASCII          3805
#define IDS_PASSWORD_TOO_LONG           3806

#define IDT_COMPRESS_ARCHIVE            4001
#define IDT_COMPRESS_UPDATE_MODE        4002
#define IDT_COMPRESS_FORMAT             4003
#define IDT_COMPRESS_LEVEL              4004
#define IDT_COMPRESS_METHOD             4005
#define IDT_COMPRESS_DICTIONARY         4006
#define IDT_COMPRESS_ORDER              4007
#define IDT_COMPRESS_SOLID              4008
#define IDT_COMPRESS_THREADS            4009
#define IDT_COMPRESS_PARAMETERS         4010
#define IDG_COMPRESS_OPTIONS            4011

#define IDX_COMPRESS_SFX                4012
#define IDX_COMPRESS_SHARED             4013

#define IDG_COMPRESS_ENCRYPTION         4014
#define IDT_COMPRESS_ENCRYPTION_METHOD  4015
#define IDX_COMPRESS_ENCRYPT_FILE_NAMES 4016
                                         
#define IDT_COMPRESS_MEMORY             4017
#define IDT_COMPRESS_MEMORY_DE          4018

#define IDX_COMPRESS_DEL                4019

#define IDX_COMPRESS_NT_SYM_LINKS       4040
#define IDX_COMPRESS_NT_HARD_LINKS      4041
#define IDX_COMPRESS_NT_ALT_STREAMS     4042
#define IDX_COMPRESS_NT_SECUR           4043

#define IDS_METHOD_STORE                4050
#define IDS_METHOD_FASTEST              4051
#define IDS_METHOD_FAST                 4052
#define IDS_METHOD_NORMAL               4053
#define IDS_METHOD_MAXIMUM              4054
#define IDS_METHOD_ULTRA                4055

#define IDS_COMPRESS_UPDATE_MODE_ADD    4060
#define IDS_COMPRESS_UPDATE_MODE_UPDATE 4061
#define IDS_COMPRESS_UPDATE_MODE_FRESH  4062
#define IDS_COMPRESS_UPDATE_MODE_SYNC   4063

#define IDS_COMPRESS_SET_ARCHIVE_BROWSE 4070
#define IDS_OPEN_TYPE_ALL_FILES         4071
#define IDS_COMPRESS_NON_SOLID          4072
#define IDS_COMPRESS_SOLID              4073

#define IDT_SPLIT_TO_VOLUMES            7302
#define IDS_INCORRECT_VOLUME_SIZE       7307
#define IDS_SPLIT_CONFIRM               7308
