#define IDS_MEM_ERROR                   3000

#define IDS_CANNOT_CREATE_FOLDER        3003
#define IDS_UPDATE_NOT_SUPPORTED        3004
#define IDS_CANT_OPEN_ARCHIVE           3005
#define IDS_CANT_OPEN_ENCRYPTED_ARCHIVE 3006
#define IDS_UNSUPPORTED_ARCHIVE_TYPE    3007

#define IDS_CANT_OPEN_AS_TYPE           3017
#define IDS_IS_OPEN_AS_TYPE             3018
#define IDS_IS_OPEN_WITH_OFFSET         3019

#define IDS_PROGRESS_EXTRACTING         3300

#define IDS_PROGRESS_SKIPPING           3325

#define IDS_EXTRACT_SET_FOLDER          3402

#define IDS_EXTRACT_PATHS_FULL          3411
#define IDS_EXTRACT_PATHS_NO            3412
#define IDS_EXTRACT_PATHS_ABS           3413
#define IDS_PATH_MODE_RELAT             3414

#define IDS_EXTRACT_OVERWRITE_ASK             3421
#define IDS_EXTRACT_OVERWRITE_WITHOUT_PROMPT  3422
#define IDS_EXTRACT_OVERWRITE_SKIP_EXISTING   3423
#define IDS_EXTRACT_OVERWRITE_RENAME          3424
#define IDS_EXTRACT_OVERWRITE_RENAME_EXISTING 3425

#define IDS_EXTRACT_MESSAGE_UNSUPPORTED_METHOD    3700
#define IDS_EXTRACT_MESSAGE_DATA_ERROR            3701
#define IDS_EXTRACT_MESSAGE_CRC_ERROR             3702
#define IDS_EXTRACT_MESSAGE_DATA_ERROR_ENCRYPTED  3703
#define IDS_EXTRACT_MESSAGE_CRC_ERROR_ENCRYPTED   3704

#define IDS_EXTRACT_MSG_WRONG_PSW_GUESS      3710
// #define IDS_EXTRACT_MSG_ENCRYPTED            3711

#define IDS_EXTRACT_MSG_UNSUPPORTED_METHOD   3721
#define IDS_EXTRACT_MSG_DATA_ERROR           3722
#define IDS_EXTRACT_MSG_CRC_ERROR            3723
#define IDS_EXTRACT_MSG_UNAVAILABLE_DATA     3724
#define IDS_EXTRACT_MSG_UEXPECTED_END        3725
#define IDS_EXTRACT_MSG_DATA_AFTER_END       3726
#define IDS_EXTRACT_MSG_IS_NOT_ARC           3727
#define IDS_EXTRACT_MSG_HEADERS_ERROR        3728
#define IDS_EXTRACT_MSG_WRONG_PSW_CLAIM      3729

#define IDS_OPEN_MSG_UNAVAILABLE_START       3763
#define IDS_OPEN_MSG_UNCONFIRMED_START       3764
#define IDS_OPEN_MSG_UNSUPPORTED_FEATURE     3768
